VERSION = "0.2.1"
CODENAME = "Buccinator"
OVERALL_DEVELOPMENT_STATUS = "Beta"

P2P_TRANSPORT_STATUS = "Ready"
LUBBADUBDUB_STATUS = "Ready"
ASSETS_STATUS = "Not Implemented"
BLOCKCHAIN_STATUS = "Ready"
WALLET_STATUS = "Beta"
RPC_STATUS = "Beta"
